@@ -1,7 +1,7 @@
 // This class is both used to resolve the dynamic variable names and
 // to provide their values. These two functions can also be
 // implemented by two different classes.
-public class DVResolverProvider implements gnu.jel.DVResolver {
+public class DVResolverProvider extends gnu.jel.DVMap {
 
     private java.util.Hashtable properties=new java.util.Hashtable();
 
