@@ -1,5 +1,5 @@
 /* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
- * $Id: TestSuite.java,v 1.35 2001/11/07 22:37:07 metlov Exp $
+ * $Id: TestSuite.java,v 1.42 2002/02/12 19:44:11 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -15,10 +15,11 @@
 
 
 import gnu.jel.*;
-import gnu.jel.generated.EC;
-import gnu.jel.generated.ParseException;
-import gnu.jel.generated.TokenMgrError;
+//import gnu.jel.generated.EC;
+//import gnu.jel.generated.ParseException;
+//import gnu.jel.generated.TokenMgrError;
 import java.io.PrintStream;
+import java.util.Stack;
 
 public class TestSuite {
   
@@ -67,17 +68,17 @@ public class TestSuite {
     testError("1=2",null,lib,3,o);
     abort_at_error=old_abort_at_error;
 
-    testError("1.0-+1.0",null,lib,5,o);
+    testError("1.0-+1.0",null,lib,6,o);
     testError("1.0&1.0",null,lib,4,o);
     testError("-",null,lib,1,o);
-    testError("0x56+0xXX",null,lib,7,o);
+    testError("0x56+0xXX",null,lib,8,o);
     testError("Sin(x)",null,lib,5,o);
     testError("Sin(6)",null,lib,1,o);
     testError("sin(' ')",null,lib,1,o);
     //    testError("'a'+'b'",null,lib,4,o); // <- now perfectly legal 
     testError("1+sin(1,6)",null,lib,3,o);
-    testError("2147483649L+2147483649",null,lib,13,o);
-    testError("01234567+08+5",null,lib,11,o);
+    testError("2147483649L+2147483649",null,lib,22,o);
+    testError("01234567+08+5",null,lib,12,o);
     testError("0.5+0#4",null,lib,6,o);
     testError("0.5+1",Integer.TYPE,lib,5,o);
     testError("0.5+(floatp)0.4D",null,lib,6,o);
@@ -198,6 +199,25 @@ public class TestSuite {
                    "2.0-1.0)/2.0)==1",
                    Boolean.TRUE,null,null,lib,o);
 
+    testExpression("true?\"a\"+\"b\":\"c\"","ab",null,null,lib,o);
+
+    testExpression("true?\"a\"+\"b\":\"c\"+\"d\"","ab",null,null,lib,o);
+
+    testExpression("true?\"ab\":\"c\"+\"d\"","ab",null,null,lib,o);
+
+    testExpression("false?\"a\"+\"b\":\"c\"+\"d\"","cd",null,null,lib,o);
+
+    testExpression("false?\"ab\":\"c\"+\"d\"","cd",null,null,lib,o);
+
+    testExpression("false?\"ab\":\"cd\"","cd",null,null,lib,o);
+
+    testExpression("false?\"ab\":\"cd\"","cd",null,null,lib,o);
+    
+    testExpression("(false?\"a\"+\"b\":\"c\"+\"d\")+\"e\"","cde",
+                   null,null,lib,o);
+
+    testExpression("\"e\"+(false?\"a\"+\"b\":\"c\"+\"d\")","ecd",
+                   null,null,lib,o);
 
     // Tests if NaN (floating point "not a number") handling conforms to JLS
     testExpression("(1>NaNd)",Boolean.FALSE,null,null,lib,o);
@@ -362,7 +382,7 @@ public class TestSuite {
 
     testError("\"abc\".nomethod()",null,lib,7,o);
     testError("\"abc\".length(666)",null,lib,7,o);
-    testError("2.0.doubleValue()",null,lib,5,o);
+    testError("2.0.doubleValue()",null,lib,4,o);
     
     testExpression("intObj+1", new Integer(556), null, rtp, lib,o);
     testExpression("(int)intObj", new Integer(555), null, rtp, lib,o);
@@ -685,13 +705,11 @@ public class TestSuite {
         // determine the result type independently
         int resID=-1;
         try {
-          OPlist list=new OPlist();
-          TypesStack typeStk=new TypesStack();
-          OP o1=new OPload(typeStk,typeConstants[i]);
-          list.addLast(o1);
+          Stack paramOPs=new Stack();
+          paramOPs.push(new OPload(typeConstants[i]));
 
-          list.addLast(new OPunary(typeStk,code));
-          resID=typeStk.peekID();
+          paramOPs.push(new OPunary(paramOPs,code));
+          resID=((OP)paramOPs.peek()).resID;
           npbcActual++;
         } catch (CompilationException exc) {
         };
@@ -726,7 +744,7 @@ public class TestSuite {
           default:
             o.print("*** : ");            
             o.println("The result of unary operation \""+opSymbols[code]+
-                      "\" on "+TypesStack.specialTypes[i]+
+                      "\" on "+OP.specialTypes[i]+
                       " is not primitive type ID="+resID);
             FAIL(o);
           };
@@ -765,8 +783,10 @@ public class TestSuite {
     String[] typeNames={"Boolean","Byte","Character","Short","Integer",
                         "Long","Float","Double"};
 
-    String[] opSymbols=(String[])TableKeeper.getTable("binOpSymbols");
-    
+    String[] opSymbols={
+        "+","-","*","/","%","&","|","^","==","!=","<",">=",
+        ">","<=","<<",">>",">>>","&&","||","{}",".+."};
+
     Object[] typeConstants={
       new java.lang.Boolean(true),
       new java.lang.Byte((byte)1),
@@ -784,19 +804,12 @@ public class TestSuite {
         // determine the result type independently
         int resID=-1;
         try {
-          OPlist list=new OPlist();
-          TypesStack typeStk=new TypesStack();
-          java.util.Stack paramOPs=new java.util.Stack();
-          OP o1=new OPload(typeStk,typeConstants[i]);
-          list.addLast(o1);
-          paramOPs.push(o1);
+          Stack paramOPs=new Stack();
+          paramOPs.push(new OPload(typeConstants[i]));
+          paramOPs.push(new OPload(typeConstants[j]));
 
-          OP o2=new OPload(typeStk,typeConstants[j]);
-          list.addLast(o2);
-          paramOPs.push(o2);
-
-          list.addLast(new OPbinary(typeStk,paramOPs,code,list));
-          resID=typeStk.peekID();
+          paramOPs.push(new OPbinary(paramOPs,code));
+          resID=((OP)paramOPs.peek()).resID;
           npbcActual++;
         } catch (CompilationException exc) {
         };
@@ -831,8 +844,8 @@ public class TestSuite {
           default:
             o.print("*** : ");            
             o.println("The result of binary operation \""+opSymbols[code]+
-                      "\" on "+TypesStack.specialTypes[i]+" and "+
-                      TypesStack.specialTypes[j]+
+                      "\" on "+OP.specialTypes[i]+" and "+
+                      OP.specialTypes[j]+
                       " is not primitive type ID="+resID);
             FAIL(o);
           };
@@ -897,7 +910,7 @@ public class TestSuite {
       // Now we need to calculate cexpr
       
       // First parse it
-      OPlist op=null;
+      OP op=null;
       try {
         op=OpenEvaluator.parse(cexpr.toString(),lib,null);
       } catch (CompilationException ce) {
@@ -947,7 +960,11 @@ public class TestSuite {
         };
         result=result1;
 
-        if (iteration==0) op.performCF();
+        if (iteration==0) 
+          try {
+            op=new OPload(op,op.eval());
+          } catch (Exception exc) {
+          };
       };
       
       if (showcases) {
@@ -968,7 +985,7 @@ public class TestSuite {
     } else {
       o.println("\"   Should throw an exception at run time.");
     };
-    OPlist op=null;
+    OP op=null;
     try {
       op=OpenEvaluator.parse(expr,lib,fixType);
     } catch (CompilationException ce) {
@@ -987,7 +1004,7 @@ public class TestSuite {
     boolean testok=true;
     
     for(int iteration=0;iteration<2;iteration++) {
-      String message=""+iteration+" |"+op.toString();
+      String message=""+iteration+" |"+toStr(op);
       o.print(message);
       for (int k=message.length();k<59;k++) o.print(' ');
 
@@ -1023,8 +1040,12 @@ public class TestSuite {
         testok=(result==null);
         if (result!=null) o.println(" ="+result.toString());
       };
-      
-      if (iteration==0) op.performCF();
+
+      if (iteration==0) 
+        try {
+          op=new OPload(op,op.eval());
+        } catch (Exception exc) {
+        };
     };
     
     if (testok) OK(o); else FAIL(o);
@@ -1037,7 +1058,7 @@ public class TestSuite {
 
     CompilationException ce=null;
     try {
-      OPlist op=OpenEvaluator.parse(expr,lib,fixType);
+      OP op=OpenEvaluator.parse(expr,lib,fixType);
     } catch (CompilationException e) {
       ce=e;
     };
@@ -1072,18 +1093,73 @@ public class TestSuite {
     successes++;
     o.println(" TEST PASSED."); //o.println();
   };
-  
+
+  public static String toStr(OP o) {
+    if (o instanceof OPload) {
+      OPload op=(OPload)o;
+      if (op.resID==8) return "\""+op.what+"\"";
+      return op.what.toString()+(op.resID>9?'L':"ZBCSIJFDLV".charAt(op.resID));
+    };
+    if (o instanceof OPbinary) {
+      String[] opSymbols={
+        "+","-","*","/","%","&","|","^","==","!=","<",">=",
+        ">","<=","<<",">>",">>>","&&","||","{}",".+."};
+      OPbinary op=(OPbinary)o;
+      return toStr(op.chi[0])+opSymbols[op.code]+toStr(op.chi[1]);
+    };
+    if (o instanceof OPunary) {
+      String[] opSymbols={"--","~","!","<RET>","(Z)","(B)",
+                          "(C)","(S)","(I)","(J)",
+                          "(F)","(D)","(L)","(POP)","->TSB","->STR"};
+      OPunary op=(OPunary)o;
+      return opSymbols[op.code]+toStr(op.chi[0]);      
+    };
+    if (o instanceof OPcall) {
+      OPcall op=(OPcall)o;
+      if (op.m==null)
+        return "{"+op.nplv+"}";
+      else {
+        StringBuffer res=new StringBuffer(op.m.getName());
+        res.append('(');
+        for (int i=0;i<op.chi.length;i++) {
+          if (i>0) res.append(",");
+          res.append(toStr(op.chi[i]));
+        };
+        res.append(')');
+        return res.toString();
+      };
+    };
+    if (o instanceof OPcondtnl) {
+      OPcondtnl op=(OPcondtnl)o;
+      StringBuffer res=new StringBuffer();
+      if (op.chi[1]!=null)
+        res.append('(');
+      
+      res.append(toStr(op.chi[0]));
+      
+      if (op.chi[1]!=null) {
+        res.append('?');
+        res.append(toStr(op.chi[1]));
+        res.append(':');
+        res.append(toStr(op.chi[2]));
+        res.append(')');
+      }
+      return res.toString();
+    };
+    return "<<<<<OP TYPE NOT IDENTIFIED>>>>";
+  };
+
 };
 
 class OpenEvaluator extends Evaluator {
   
-  public static OPlist parse(String expression, Library lib,
-                             Class resultType) throws CompilationException {
+  public static OP parse(String expression, Library lib,
+                         Class resultType) throws CompilationException {
 	return Evaluator.parse(expression,lib,resultType);
   };
 
-  public static CompiledExpression compile(OPlist list) {
-	byte[] image=Evaluator.getImage(list);
+  public static CompiledExpression compile(OP op) {
+	byte[] image=Evaluator.getImage(op);
     try {
       java.io.FileOutputStream fos=new java.io.FileOutputStream("dump.class");
       fos.write(image);
