@@ -1,5 +1,5 @@
 /* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
- * $Id: Evaluator.java,v 1.14 2001/08/23 14:30:21 metlov Exp $
+ * $Id: Evaluator.java,v 1.19 2002/02/12 19:44:11 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -14,8 +14,8 @@
 
 package gnu.jel; 
 
-import gnu.jel.generated.EC;
-import gnu.jel.generated.CharStream;
+//import gnu.jel.generated.EC;
+//import gnu.jel.generated.CharStream;
 import gnu.jel.debug.Debug;
 import java.lang.reflect.Member;
 
@@ -64,7 +64,7 @@ public class Evaluator {
       paramsE[0]=(new Object[0]).getClass();
       for(int i=0;i<9;i++) {
         String name="evaluate";
-        Class cls=TypesStack.specialTypes[i];
+        Class cls=OP.specialTypes[i];
         if (i!=8) 
           name=name+'_'+cls;
         else 
@@ -79,23 +79,21 @@ public class Evaluator {
         new LocalMethod(0x0001,Void.TYPE,"<init>",null,null);
       cf.newMethod(cnstr,null);
       cf.code(0x2a);                //| aload_0  ;loads "this"
-      cf.typesStk.pushID(11); // not important what, it must be a reference
-      Member supInit=cmplExpr.getConstructor(new Class[0]);
-      cf.code(0xb7);                //| invokespecial
-      cf.codeI(cf.getIndex(supInit,10));  //|    super();
-      cf.typesStk.pop();
+      cf.noteStk(-1,11); // not important what, it must be a reference
+      cf.codeM(cmplExpr.getConstructor(new Class[0]));
+      cf.noteStk(11,-1);
       cf.code(0xb1);                //| return void
       
       LocalMethod getType=
       new LocalMethod(0x0001,Integer.TYPE,"getType",null,null);
       cf.newMethod(getType,null);
       cf.code(0x10);                //| bipush
-      retID_patchback=cf.textData.size();
+      retID_patchback=cf.tsize;
       cf.code(8);                   //    type placeholder
-      cf.typesStk.push(Integer.TYPE);
+      cf.noteStk(-1,4); // note "int"
       
       cf.code(0xAC);                //| ireturn
-      cf.typesStk.pop();
+      cf.noteStk(4,-1); // rm "int"
       
       cf_orig=(ClassFile)cf.clone();
     } catch (Exception exc) {
@@ -145,10 +143,14 @@ public class Evaluator {
   public static byte[] compileBits(String expression, Library lib,
                                    Class resultType) 
     throws CompilationException {
-    OPlist code=parse(expression,lib,resultType);
-    code.performCF(); // fold constants
-    //    System.out.println("CF performed.");
-    //    System.out.println(code);
+    OP code=parse(expression,lib,resultType);
+    
+    // Perform constants folding
+    try {
+      code=new OPload(code,code.eval());
+    } catch (Exception exc) {
+    };
+
     return getImage(code);
   };
 
@@ -184,22 +186,26 @@ public class Evaluator {
     return compileBits(expression, lib, null);
   };
 
-  protected static OPlist parse(String expression, Library lib,
+  protected static OP parse(String expression, Library lib,
                                 Class resultType) throws CompilationException {
-    return (new EC(new CharStream(expression))).parse(resultType,lib);
+    return (new Parser(expression,lib)).parse(resultType);
+    //    return (new EC(new CharStream(expression))).parse(resultType,lib);
   };
 
-  protected static byte[] getImage(OPlist list) {
-    int retID=((OPunary)list.getLast()).resID;
+  protected static byte[] getImage(OP code) {
+    int retID=code.resID;
 
     ClassFile cf=(ClassFile)cf_orig.clone();
 
     // set return type
-    cf.textData.patch(retID_patchback,(byte)retID);
+    int otsize=cf.tsize;
+    cf.tsize=retID_patchback;
+    cf.write((byte)retID);
+    cf.tsize=otsize;
 
     // add the evaluate method
     cf.newMethod(eval_methods[retID],null);
-    list.compile(cf);
+    code.compile(cf);
 
     return cf.getImage();
   };
