@@ -1,5 +1,5 @@
 /*-*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
- * $Id: Library.java,v 1.15 2001/11/07 22:38:21 metlov Exp $
+ * $Id: Library.java,v 1.19 2002/02/12 19:44:11 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -18,12 +18,13 @@ import gnu.jel.debug.Debug;
 import java.lang.reflect.Member;
 import java.lang.reflect.Field;
 import java.lang.reflect.Method;
+import java.lang.reflect.Constructor;
 import java.util.Hashtable;
 import java.util.Vector;
 import java.util.Enumeration;
 
 //+-+-+-+TESTS
-//-import gnu.jel.debug.Tester;
+import gnu.jel.debug.Tester;
 //+-+-+-+TESTS
 
 /**
@@ -220,12 +221,11 @@ public class Library {
 
   private static boolean rehash(Hashtable hashedNames, Member m) {
     String name=m.getName();
-    String signature=TypesStack.getSignature(m);
+    String signature=getSignature(m);
 
     // for the purpose of matching fields behave like methods with no
     // arguments
-    if ((m instanceof Field) || (m instanceof LocalField))
-      signature="()"+signature;
+    if (isField(m)) signature="()"+signature;
     
     Hashtable signatures=(Hashtable)hashedNames.get(name);
     if (signatures==null) { 
@@ -312,11 +312,11 @@ public class Library {
     // dot operator security    
     if (container!=null) {
       if (dotClasses==null) // dot operator is prohibited
-        throw new CompilationException(-1,11,null);
+        throw new CompilationException(11,null);
       else if (! (noDotSecurity || dotClasses.containsKey(container))) {
         // dot is not allowed in this particular class
         Object[] paramsExc={container};
-        throw new CompilationException(-1,12,paramsExc);
+        throw new CompilationException(12,paramsExc);
       };
       if ((hashedMembers=(Hashtable)dotClasses.get(container))==null) {
         rehash(container);
@@ -332,21 +332,21 @@ public class Library {
 
     if (signatures==null) { // name is not found
       Object[] paramsExc={name,container};
-      throw new CompilationException(-1,container==null?5:6,paramsExc);
+      throw new CompilationException(container==null?5:6,paramsExc);
     };
     
     // Choose applicable methods
     Vector applicable_methods=new Vector();
     for(Enumeration e=signatures.elements();e.hasMoreElements();) {
       Member cm = (Member)e.nextElement();
-      Class[] cp=TypesStack.getParameterTypes(cm);
+      Class[] cp=getParameterTypes(cm);
       
       boolean applicable=false;
       if (params!=null) { // if no parameters requested
         if (cp.length==params.length) {  // If number of arguments matches
           applicable=true;
           for(int i=0;((i<cp.length) && applicable);i++) {
-            applicable=TypesStack.isWidening(params[i],cp[i]);
+            applicable=OP.isWidening(params[i],cp[i]);
           };
         };
       } else {
@@ -358,7 +358,7 @@ public class Library {
     
     if (applicable_methods.size()==0) {
       Object[] paramsExc={name,describe(name,params),container};
-      throw new CompilationException(-1,container==null?7:8,paramsExc);
+      throw new CompilationException(container==null?7:8,paramsExc);
     };
     
     if (applicable_methods.size()==1) 
@@ -367,14 +367,14 @@ public class Library {
     // Search for the most specific method
     Enumeration e=applicable_methods.elements();
     Member most_specific=(Member)e.nextElement();
-    Class[] most_specific_params=TypesStack.getParameterTypes(most_specific);
+    Class[] most_specific_params=getParameterTypes(most_specific);
 
     //    System.out.println("--- APPLICABLE METHODS ---");
     //    System.out.println(most_specific.getName()+
     //                       ClassFile.getSignature(most_specific));
     while (e.hasMoreElements()) {
       Member cm= (Member)e.nextElement();
-      Class[] cp=TypesStack.getParameterTypes(cm);
+      Class[] cp=getParameterTypes(cm);
       boolean moreSpecific=true;
       boolean lessSpecific=true;
 
@@ -382,9 +382,9 @@ public class Library {
       
       for(int i=0; i<cp.length; i++) {
         moreSpecific = moreSpecific && 
-          TypesStack.isWidening(cp[i],most_specific_params[i]);
+          OP.isWidening(cp[i],most_specific_params[i]);
         lessSpecific = lessSpecific &&
-          TypesStack.isWidening(most_specific_params[i],cp[i]);
+          OP.isWidening(most_specific_params[i],cp[i]);
       };
       
       if (moreSpecific && (!lessSpecific)) {
@@ -397,7 +397,7 @@ public class Library {
                            describe(name,cp),
                            describe(name,params),
                            container};
-        throw new CompilationException(-1,container==null?9:10,paramsExc);        
+        throw new CompilationException(container==null?9:10,paramsExc);        
       };
     };
     //    System.out.println("--- END APPLICABLE METHODS ---");
@@ -432,241 +432,348 @@ public class Library {
     if (id==null) return -1;
     return id.intValue();
   };
+
+    /**
+   * Used to get return type of a class member.
+   * <P>The type of a Method is its return type, the type of a Constructor is
+   * void.
+   * @param  m member whose type is to be determined
+   * @return type of the member
+   */
+  public static Class getType(Member m) {
+    if (m instanceof Method) return ((Method)m).getReturnType();
+    if (m instanceof Field) return ((Field)m).getType();
+    if (m instanceof LocalField) return ((LocalField)m).getType();
+    // otherwise it must be java.lang.reflect.Constructor
+    if (Debug.enabled)
+      Debug.assert(m instanceof java.lang.reflect.Constructor);
+    return OP.specialTypes[9]; // java.lang.reflect.Void.TYPE
+  };
+
+  /**
+   * Used to get types of formal parameters of a member.
+   * <P> The reference to the class instance "this" is not counted by
+   * this method.
+   * @param  m member whose formal parameters are to be obtained
+   * @return array of formal parameter types (empty array if none).
+   */
+  public static Class[] getParameterTypes(Member m) {
+    if (m instanceof Method) return ((Method)m).getParameterTypes();
+    if (m instanceof LocalMethod) return ((LocalMethod)m).getParameterTypes();
+    if (m instanceof Constructor) return ((Constructor)m).getParameterTypes();
+
+    if (Debug.enabled)
+      Debug.assert((m instanceof Field)||(m instanceof LocalField));
+    
+    return new Class[0];
+  };
+
+  /**
+   * Computes signature of the given member.
+   * @param m the member to compute the sugnature of.
+   * @return the signature.
+   */
+  public static String getSignature(Member m) {
+    StringBuffer signature=new StringBuffer();
+    if (!isField(m)) {
+      Class parameters[]=getParameterTypes(m);
+      signature.append('(');
+      for(int i=0;i<parameters.length;i++) 
+        appendSignature(signature,parameters[i]);
+      signature.append(')');
+    };
+    appendSignature(signature,getType(m));
+    return signature.toString();
+  };
+
+  public static boolean isField(Member m) {
+    return (m instanceof Field) || ((m instanceof LocalField)
+                                    && !(m instanceof LocalMethod));
+  };
+
+  /**
+   * Computes the signature of the given class.
+   * <P> The signature of the class (Field descriptor) is the string and 
+   * it's format is described in the paragraph 4.3.2 of the Java VM 
+   * specification (ISBN 0-201-63451-1).
+   * <P>The same can be done using <TT>java.lang.Class.getName()</TT> by 
+   * converting it's result into the "historical form".
+   * <P> This utility method can be used outside of the JEL package
+   * it does not involve any JEL specific assumptions and should follow
+   * JVM Specification precisely.
+   * @param cls is the class to compute the signature of. Can be primitive or
+   *            array type.
+   * @return the class signature.
+   */
+  public static String getSignature(Class cls) {
+    return appendSignature(new StringBuffer(),cls).toString();
+  };
+
+  private static StringBuffer appendSignature(StringBuffer buff, Class cls) {
+    if (cls.isPrimitive()) {
+      int tid;
+      buff.append((tid=OP.typeID(cls))>9?'L':"ZBCSIJFDLV".charAt(tid));
+    } else if (cls.isArray()) {
+      buff.append('[');
+      appendSignature(buff,cls.getComponentType());
+    } else { // just a class
+      buff.append('L');
+      appendHistoricalForm(buff,cls.getName());
+      buff.append(';');
+    };
+    return buff;
+  };
+
+  public static String toHistoricalForm(String className) {
+    return appendHistoricalForm(new StringBuffer(),className).toString();
+  };
+
+  private static StringBuffer appendHistoricalForm(StringBuffer buff,
+                                                   String className) {
+    int namelen=className.length();
+    for(int i=0;i<className.length();i++) {
+      char cch=className.charAt(i);
+      if (cch=='.') cch='/';
+      buff.append(cch);
+    };
+    return buff;
+  };
+
   
   //+-+-+-+TESTS  
-//-  // ------------------------- TESTSUITE -------------------------------
-//-  
-//-  /**
-//-   * Performs unitary test of the library.
-//-   * @param args ignored.
-//-   */
-//-  public static void main(String[] args) {
-//-    Tester t=new Tester(System.out);
-//-    test(t);
-//-    t.summarize();
-//-  };
-//-
-//-  /**
-//-   * Performs unitary test of the library.
-//-   * <p> Used if all package is being tested and not just codegen.
-//-   * @param t Tester to report test results.
-//-   */
-//-  public static void test(Tester t) {
-//-    Library ll=null;
-//-    Class math=null;
-//-    try {
-//-      math=Class.forName("java.lang.Math");
-//-    } catch(ClassNotFoundException e) {
-//-      Debug.println("It is IMPOSSIBLE :)");
-//-    };
-//-
-//-    t.startTest("Creating the library of java.lang.Math");
-//-    try {
-//-      Class[] sl=new Class[1];
-//-      sl[0]=math;
-//-      Library l=new Library(sl,null);
-//-      ll=l;
-//-      t.testOK();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Attempt of invocation round(double)");
-//-    try {
-//-      Class[] par=new Class[1];
-//-      par[0]=Double.TYPE;
-//-      Member mf=ll.getMember(null,"round",par);
-//-      if ((mf!=null) && 
-//-          (mf.equals(math.getMethod("round",par))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Attempt of invocation round(float)");
-//-    try {
-//-      Class[] par=new Class[1];
-//-      par[0]=Float.TYPE;
-//-      Member mf=ll.getMember(null,"round",par);
-//-      if ((mf!=null) && 
-//-          (mf.equals(math.getMethod("round",par))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Attempt of invocation round(int) best is round(float)");
-//-    try {
-//-      Class[] par=new Class[1];
-//-      par[0]=Integer.TYPE;
-//-      Member mf=ll.getMember(null,"round",par);
-//-
-//-      Class[] par1=new Class[1];
-//-      par1[0]=Float.TYPE;
-//-
-//-
-//-      if ((mf!=null) && 
-//-          (mf.equals(math.getMethod("round",par1))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Attempt of invocation abs(int) best is abs(int)");
-//-    try {
-//-      Class[] par=new Class[1];
-//-      par[0]=Integer.TYPE;
-//-      Member mf=ll.getMember(null,"abs",par);
-//-
-//-      Class[] par1=new Class[1];
-//-      par1[0]=Integer.TYPE;
-//-
-//-      if ((mf!=null) && 
-//-          (mf.equals(math.getMethod("abs",par1))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Attempt of invocation abs(byte) best is abs(int)");
-//-    try {
-//-      Class[] par=new Class[1];
-//-      par[0]=Byte.TYPE;
-//-      Member mf=ll.getMember(null,"abs",par);
-//-      
-//-      Class[] par1=new Class[1];
-//-      par1[0]=Integer.TYPE;
-//-      
-//-      if ((mf!=null) && 
-//-          (mf.equals(math.getMethod("abs",par1))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Attempt of invocation abs(char) best is abs(int)");
-//-    try {
-//-      Class[] par=new Class[1];
-//-      par[0]=Character.TYPE;
-//-      Member mf=ll.getMember(null,"abs",par);
-//-      
-//-      Class[] par1=new Class[1];
-//-      par1[0]=Integer.TYPE;
-//-      
-//-      if ((mf!=null) && 
-//-          (mf.equals(math.getMethod("abs",par1))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-    
-//-    t.startTest("Attempt of invocation min(int,float) best is min(float,float)");
-//-    try {
-//-      Class[] par=new Class[2];
-//-      par[0]=Integer.TYPE;
-//-      par[1]=Float.TYPE;
-//-      Member mf=ll.getMember(null,"min",par);
-//-      
-//-      Class[] par1=new Class[2];
-//-      par1[0]=Float.TYPE;
-//-      par1[1]=Float.TYPE;
-//-      
-//-      if ((mf!=null) && 
-//-          (mf.equals(math.getMethod("min",par1))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Attempt of access to the field PI");
-//-    try {
-//-      Class[] par=new Class[0];
-//-      Member f=ll.getMember(null,"PI",par);
-//-      if ((f!=null) && (f.getName().equals("PI")))
-//-	t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Checking assignment of state dependence ");
-//-    try {
-//-      ll.markStateDependent("random",null);
-//-      if (!ll.isStateless(ll.getMember(null,"random",null)))
-//-	t.testOK();
-//-      else
-//-	t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Creating a library with dot operations allowed ");
-//-    Library ldot=null;
-//-    try {
-//-      Class[] stl=new Class[1];
-//-      Class[] dynl=new Class[1];
-//-      Class[] dotl=new Class[2];
-//-      stl[0]=Class.forName("java.lang.Math");
-//-      dynl[0]=Class.forName("java.util.Random");
-//-      dotl[0]=Class.forName("java.util.Hashtable");
-//-      dotl[1]=Class.forName("java.util.Vector");
-//-      
-//-      ldot=new Library(stl,dynl,dotl);
-//-      t.testOK();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Resolving size() on Hashtable");
-//-    try {
-//-      Class[] params=new Class[0];
-//-      Class htable=Class.forName("java.util.Hashtable");
-//-      Member mf=ldot.getMember(htable,"size",params);
-//-      if ((mf!=null) && 
-//-          (mf.equals(htable.getMethod("size",params))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Resolving size() on Vector");
-//-    try {
-//-      Class[] params=new Class[0];
-//-      Class vctr=Class.forName("java.util.Vector");
-//-      Member mf=ldot.getMember(vctr,"size",params);
-//-      if ((mf!=null) && 
-//-          (mf.equals(vctr.getMethod("size",params))))
-//-	  t.testOK(); else t.testFail();
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-    t.startTest("Ensuring size() is not leaked to global context.");
-//-    try {
-//-      Class[] params=new Class[0];
-//-      try {
-//-        Member mf=ldot.getMember(null,"size",params);
-//-        t.testFail();
-//-      } catch (CompilationException exc) {
-//-        t.testOK();
-//-      };
-//-    } catch (Throwable e) {
-//-      Debug.reportThrowable(e);
-//-      t.testFail();
-//-    };
-//-
-//-  };
+  // ------------------------- TESTSUITE -------------------------------
+  
+  /**
+   * Performs unitary test of the library.
+   * @param args ignored.
+   */
+  public static void main(String[] args) {
+    Tester t=new Tester(System.out);
+    test(t);
+    t.summarize();
+  };
+
+  /**
+   * Performs unitary test of the library.
+   * <p> Used if all package is being tested and not just codegen.
+   * @param t Tester to report test results.
+   */
+  public static void test(Tester t) {
+    Library ll=null;
+    Class math=null;
+    try {
+      math=Class.forName("java.lang.Math");
+    } catch(ClassNotFoundException e) {
+      Debug.println("It is IMPOSSIBLE :)");
+    };
+
+    t.startTest("Creating the library of java.lang.Math");
+    try {
+      Class[] sl=new Class[1];
+      sl[0]=math;
+      Library l=new Library(sl,null);
+      ll=l;
+      t.testOK();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Attempt of invocation round(double)");
+    try {
+      Class[] par=new Class[1];
+      par[0]=Double.TYPE;
+      Member mf=ll.getMember(null,"round",par);
+      if ((mf!=null) && 
+          (mf.equals(math.getMethod("round",par))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Attempt of invocation round(float)");
+    try {
+      Class[] par=new Class[1];
+      par[0]=Float.TYPE;
+      Member mf=ll.getMember(null,"round",par);
+      if ((mf!=null) && 
+          (mf.equals(math.getMethod("round",par))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Attempt of invocation round(int) best is round(float)");
+    try {
+      Class[] par=new Class[1];
+      par[0]=Integer.TYPE;
+      Member mf=ll.getMember(null,"round",par);
+
+      Class[] par1=new Class[1];
+      par1[0]=Float.TYPE;
+
+
+      if ((mf!=null) && 
+          (mf.equals(math.getMethod("round",par1))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Attempt of invocation abs(int) best is abs(int)");
+    try {
+      Class[] par=new Class[1];
+      par[0]=Integer.TYPE;
+      Member mf=ll.getMember(null,"abs",par);
+
+      Class[] par1=new Class[1];
+      par1[0]=Integer.TYPE;
+
+      if ((mf!=null) && 
+          (mf.equals(math.getMethod("abs",par1))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Attempt of invocation abs(byte) best is abs(int)");
+    try {
+      Class[] par=new Class[1];
+      par[0]=Byte.TYPE;
+      Member mf=ll.getMember(null,"abs",par);
+      
+      Class[] par1=new Class[1];
+      par1[0]=Integer.TYPE;
+      
+      if ((mf!=null) && 
+          (mf.equals(math.getMethod("abs",par1))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Attempt of invocation abs(char) best is abs(int)");
+    try {
+      Class[] par=new Class[1];
+      par[0]=Character.TYPE;
+      Member mf=ll.getMember(null,"abs",par);
+      
+      Class[] par1=new Class[1];
+      par1[0]=Integer.TYPE;
+      
+      if ((mf!=null) && 
+          (mf.equals(math.getMethod("abs",par1))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+    
+    t.startTest("Attempt of invocation min(int,float) best is min(float,float)");
+    try {
+      Class[] par=new Class[2];
+      par[0]=Integer.TYPE;
+      par[1]=Float.TYPE;
+      Member mf=ll.getMember(null,"min",par);
+      
+      Class[] par1=new Class[2];
+      par1[0]=Float.TYPE;
+      par1[1]=Float.TYPE;
+      
+      if ((mf!=null) && 
+          (mf.equals(math.getMethod("min",par1))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Attempt of access to the field PI");
+    try {
+      Class[] par=new Class[0];
+      Member f=ll.getMember(null,"PI",par);
+      if ((f!=null) && (f.getName().equals("PI")))
+	t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Checking assignment of state dependence ");
+    try {
+      ll.markStateDependent("random",null);
+      if (!ll.isStateless(ll.getMember(null,"random",null)))
+	t.testOK();
+      else
+	t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Creating a library with dot operations allowed ");
+    Library ldot=null;
+    try {
+      Class[] stl=new Class[1];
+      Class[] dynl=new Class[1];
+      Class[] dotl=new Class[2];
+      stl[0]=Class.forName("java.lang.Math");
+      dynl[0]=Class.forName("java.util.Random");
+      dotl[0]=Class.forName("java.util.Hashtable");
+      dotl[1]=Class.forName("java.util.Vector");
+      
+      ldot=new Library(stl,dynl,dotl);
+      t.testOK();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Resolving size() on Hashtable");
+    try {
+      Class[] params=new Class[0];
+      Class htable=Class.forName("java.util.Hashtable");
+      Member mf=ldot.getMember(htable,"size",params);
+      if ((mf!=null) && 
+          (mf.equals(htable.getMethod("size",params))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Resolving size() on Vector");
+    try {
+      Class[] params=new Class[0];
+      Class vctr=Class.forName("java.util.Vector");
+      Member mf=ldot.getMember(vctr,"size",params);
+      if ((mf!=null) && 
+          (mf.equals(vctr.getMethod("size",params))))
+	  t.testOK(); else t.testFail();
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+    t.startTest("Ensuring size() is not leaked to global context.");
+    try {
+      Class[] params=new Class[0];
+      try {
+        Member mf=ldot.getMember(null,"size",params);
+        t.testFail();
+      } catch (CompilationException exc) {
+        t.testOK();
+      };
+    } catch (Throwable e) {
+      Debug.reportThrowable(e);
+      t.testFail();
+    };
+
+  };
   //+-+-+-+TESTS
 };
