@@ -1,5 +1,5 @@
 /* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- 
- * $Id: OPbinary.java,v 1.9 2001/08/23 14:30:21 metlov Exp $
+ * $Id: OPbinary.java,v 1.19 2002/02/13 02:14:44 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -17,36 +17,34 @@ package gnu.jel;
 import gnu.jel.debug.Debug;
 import java.util.Stack;
 
-public class OPbinary extends OPfunction {
-  private int code;
+/**
+ * A tree node, representing binary operation.
+ */
+public class OPbinary extends OP {
 
-  private int opsIDX; // index into ops array to get code for this op.
+  /** code of this operation */
+  public int code;
+
+  /** index into ops array to get code for this op. */
+  private int opsIDX; 
 
   // Names of binary operations by ID in the readable form.
   private final static String[] opNames;
 
-  // Symbols of binary operations by ID in the readable form.
-  private final static String[] opSymbols;
-
-  // binary promotions of base types
+  /** binary promotions of base types */
   protected final static byte[][] promotions;
 
   // code chunks to implement the ops
   private final static  int[][] ops;
   
-  // open jumps left after the ops
-  private static final char[][] openjumps;
-
   // type of operands promotion by the opcode
   private static final byte[] promotionTypes;
 
   static {
     promotions=(byte[][])TableKeeper.getTable("promotions");
     ops=(int[][])TableKeeper.getTable("ops");
-    openjumps=(char[][])TableKeeper.getTable("openjumps");
     promotionTypes=(byte[])TableKeeper.getTable("promotionTypes");
     opNames=(String[])TableKeeper.getTable("binOpNames");
-    opSymbols=(String[])TableKeeper.getTable("binOpSymbols");
   };
 
   /**
@@ -73,27 +71,27 @@ public class OPbinary extends OPfunction {
    * 17  --  logical conjunction operator (AND)
    * 18  --  logical disjunction operator (OR)
    * 19  --  array element access operation
-   * 20  --  reserved (used internally for sctring concatenation)
+   * 20  --  reserved (used internally for string concatenation)
    * </PRE>
-   * @param typesStk holds current items on stack
+   * @param paramOPs stack holding the operands
    * @param opcode is the operation code
    */
-  public OPbinary(TypesStack typesStk, Stack paramOPs, int opcode,OPlist list) 
-    throws CompilationException {
+  public OPbinary(Stack paramOPs, int opcode) throws CompilationException {
     if (Debug.enabled)
-      Debug.assert((opcode>=0) && (opcode<opNames.length) && 
-                   (opNames.length==opSymbols.length));
+      Debug.assert((opcode>=0) && (opcode<opNames.length));
 
     // get the operands types
-    int op2ID=typesStk.peekID();
-    Class op2Type=typesStk.pop();
-    int op1ID=typesStk.peekID();
-    Class op1Type=typesStk.pop();
-    OP p2CvtPlaceholder=(OP)paramOPs.pop();
-    OP p1CvtPlaceholder=(OP)paramOPs.pop();
+    chi=new OP[2];
+    chi[1]=(OP)paramOPs.pop();
+    chi[0]=(OP)paramOPs.pop();
+
+    int op2ID=chi[1].resID;
+    Class op2Type=chi[1].resType;
+    int op1ID=chi[0].resID;
+    Class op1Type=chi[0].resType;
 
     // separate out string concatenation
-    if (((op1ID==10) || (TypesStack.unwrapType[op1ID]==11)) && (opcode==0))
+    if (((op1ID==10) || (unwrapType[op1ID]==11)) && (opcode==0))
       opcode=20;
 
     this.code=opcode;
@@ -106,52 +104,47 @@ public class OPbinary extends OPfunction {
     
     resID=-1;
     boolean second_narrowing=false;
+    int op1IDuwrp=unwrapType[op1ID];
+    int op2IDuwrp=unwrapType[op2ID];
+
     switch (promotionTypes[opcode]) {
     case 0: // binary promotion with boolean result
       resID=0;
     case 1: // binary promotion
       op1cvtID=op2cvtID=opsIDX=
-        promotions[TypesStack.unwrapType[op1ID]][TypesStack.unwrapType[op2ID]];
-      if (op1ID==10) op1cvtID=11; // remove TSB
-      if (op2ID==10) op2cvtID=11; // remove TSB
+        promotions[op1IDuwrp][op2IDuwrp];
       if (resID<0) resID=opsIDX;
 
       if (opsIDX==-1) { // types are incompatible (can't promote)
         Object[] paramsExc={op1Type,op2Type,opNames[opcode]};
-        throw new CompilationException(-1,15,paramsExc);
+        throw new CompilationException(15,paramsExc);
       };
 
-      break;
-    case 2: // unary promotion of the first operand, second to int
-      resID=op1cvtID=opsIDX=
-        OPunary.unary_prmtns[TypesStack.unwrapType[op1ID]];
-      if (!TypesStack.isIntegral(TypesStack.unwrapType[op2ID])) {
-        Object[] paramsExc={opNames[opcode],op2Type};
-        throw new CompilationException(-1,27,paramsExc);
-      };
-      op2cvtID=4;
-      second_narrowing=true;
       break;
     case 3: // array promotion
       resType=(op1Type!=null?op1Type.getComponentType():null);
       if (resType==null)
-        throw new CompilationException(-1,18,null);
-      resID=TypesStack.typeID(resType);
-
-      if (!TypesStack.isIntegral(TypesStack.unwrapType[op2ID])) {
-      Object[] paramsExc={opNames[opcode],op2Type};
-        throw new CompilationException(-1,27,paramsExc);
+        throw new CompilationException(18,null);
+      resID=typeID(resType);
+      opsIDX=(resID>=8?8:resID);
+    case 2: // unary promotion of the first operand, second to int
+      if (resID<0) {
+        resID=op1cvtID=opsIDX=
+          OPunary.unary_prmtns[op1IDuwrp];
+        second_narrowing=true;
       };
 
+      // integral types mask
+      //      7 6 5 4 3 2 1 0
+      //      0 0 1 1 1 1 1 0 = 0x3E
+      if (((0x3E>>op2IDuwrp) & 1)==0) { // type is not integral
+        Object[] paramsExc={opNames[opcode],op2Type};
+        throw new CompilationException(27,paramsExc);
+      };
       op2cvtID=4;
-      opsIDX=TypesStack.baseType[resID];
       break;
     case 4: // string concatenation promotion
-      if (op2ID==11)
-        opsIDX=11;  // Strings handled in a special way for performance
-      else 
-        opsIDX=TypesStack.baseType[op2ID];
-
+      opsIDX=(op2ID>11?8:op2ID);
       op1cvtID=10; // TSB
       resID=10;  // TSB
       break;
@@ -164,93 +157,99 @@ public class OPbinary extends OPfunction {
     // check if the OP can be implemented
     if (ops[opcode][opsIDX]==0xFF) { // operation is not defined on types
       Object[] paramsExc={opNames[opcode],op1Type,op2Type};
-      throw new CompilationException(-1,16,paramsExc);
+      throw new CompilationException(16,paramsExc);
     };
     
     // insert type conversion opcode
     if ((op1ID!=op1cvtID) && (op1cvtID!=8)) {
-      typesStk.pushID(op1ID);
-      list.addAfter(p1CvtPlaceholder,
-                    new OPunary(typesStk,op1cvtID,null,false));
-      typesStk.pop();
+      paramOPs.push(chi[0]);
+      chi[0]=new OPunary(paramOPs,op1cvtID,null,op1cvtID==10);
+      //                                        can narrow to TSB
     };
+
     if ((op2ID!=op2cvtID) && (op2cvtID!=8)) {
-      typesStk.pushID(op2ID);
-      list.addAfter(p2CvtPlaceholder,
-                    new OPunary(typesStk,op2cvtID,null,second_narrowing));
-      typesStk.pop();
+      paramOPs.push(chi[1]);
+      chi[1]=new OPunary(paramOPs,op2cvtID,null,second_narrowing);
     };
-    
-    typesStk.pushID(resID,resType); // push the result type on stack
+
+    if (resID!=8)
+      resType=specialTypes[resID];
   };
  
-  /**
-   * Returns number of parameters for this function.
-   */
-  public int getNParams() {
-    return 2;
-  };
-  
-  protected void compile_par(ClassFile cf, int n) {
+  public void compile(ClassFile cf) {
     if ((code==17) || (code==18)) {
-      // logical AND/OR
-      if (n==1) 
-        cf.logical_param(code==17);
-      else
-        cf.logical_end(code==17);
-    } else cf.ensure_value();
-  };
+      chi[0].compile(cf);
+      cf.code(code-(17 - 0xF4)); // logical_param AND/OR
+      chi[1].compile(cf);
+      cf.code(code-(17 - 0xF6)); // logical_end AND/OR
+    } else {
+      chi[0].compile(cf);
+      cf.code(0xFA); //   ensure value;
+      chi[1].compile(cf);
+      cf.code(0xFA); //   ensure value;
 
-  protected void compile(ClassFile cf) {
-    if ((code!=17) && (code!=18)) {
-      cf.code(ops[code][opsIDX]);
-      cf.currJump=openjumps[code][opsIDX];
+      cf.code(ops[code][opsIDX] & 0xFFFFFFFFL); 
+      // & is needed to prevent sign extension when converting "int" ops
+      // element into "long" argument of code.
+      
+      cf.noteStk(chi[0].resID,-1);
+      cf.noteStk(chi[1].resID,-1);
       
-      cf.typesStk.pop();
-      cf.typesStk.pop();
-        
       if (cf.currJump==0) // jumps do not load anything to the stack
-        cf.typesStk.pushID(resID,resType);
+        cf.noteStk(-1,resID);
     };
   };
-  
-  /**
-   * Attempts to perform this operation.
-   * @param list is the list of OPs this one belong to, 
-   *             if eval is unsuccessful this list is not modified.
-   */
-  protected void eval(OPlist list) {
-    if (code==19) return; // array access can't be evaluated.
+
+  public Object eval() throws Exception {
+    
+    Object c1w=null;
+    boolean ep1=false;
+    int c1ID=chi[0].resID;
+    Object c2w=null;
+    boolean ep2=false;
+    int c2ID=chi[1].resID;
+    
     try {
-      OPload c2=(OPload) this.prev;
-      OPload c1=(OPload) this.prev.prev;
+      c1w=chi[0].eval();
+    } catch (Exception e) {
+      ep1=true;
+    };
+    
+    try {
+      c2w=chi[1].eval();
+    } catch (Exception e) {
+      ep2=true;
+    };
+
+    try {
+      if (ep1 || ep2 || (code==19)) // array access can't be evaluated.
+        throw new Exception();
+      
       if (code==20) {
-        ((StringBuffer)c1.what).append(String.valueOf(c2.what));
-      } else if ((TypesStack.baseType[c1.resID]==8) || 
-                 (TypesStack.baseType[c2.resID]==8)) {
+        ((StringBuffer)c1w).append(String.valueOf(c2w));
+      } else if ((c1ID>=8) || (c2ID>=8)) {
         // the only way the objects appear in this context
         // are the comparisons
         if (Debug.enabled) {
           Debug.assert((code>=8)||(code<=13),
                        "only comparisons and concatenation binops can "+
                        "operate on objects.");
-          Debug.assert(c1.resID!=10, "No TSB in this context 1");
-          Debug.assert(c1.resID!=10, "No TSB in this context 2");
+          Debug.assert(c1ID!=10, "No TSB in this context 1");
+          Debug.assert(c1ID!=10, "No TSB in this context 2");
         };
 
         // only string literal comparisons are interpreted
-        if ((c1.resID!=11) || (c2.resID!=11)) return;
+        if ((c1ID!=11) || (c2ID!=11))
+          throw new Exception();
 
         // compare the strings
-        int res=CompiledExpression.compare((String)c1.what,(String)c2.what);
-        c1.what=((0x3172A>>>(3*(code-8)))&((res>0)?1:((res<0)?4:2)))>0
+        int res=CompiledExpression.compare((String)c1w,(String)c2w);
+        c1w=((0x3172A>>>(3*(code-8)))&((res>0)?1:((res<0)?4:2)))>0
           ?Boolean.TRUE:Boolean.FALSE;
-        c1.resType=Boolean.TYPE;
-        c1.resID=0;
       } else {  // binary on primitive types
         // Widen
-        Number n1=TypesStack.widen(c1.what,c1.resID);
-        Number n2=TypesStack.widen(c2.what,c2.resID);
+        Number n1=widen(c1w,c1ID);
+        Number n2=widen(c2w,c2ID);
       
         // Perform      
         boolean boolres=false;
@@ -321,27 +320,25 @@ public class OPbinary extends OPfunction {
       
         // Narrow    
         if (boolres) {
-          c1.what=TypesStack.narrow(n1,0);
-          c1.resType=Boolean.TYPE;
-          c1.resID=0;
-        } else c1.what=TypesStack.narrow(n1,resID);
+          c1w=narrow(n1,0);
+        } else c1w=narrow(n1,resID);
         
       };
-      list.remove(this.prev); // remove second operand
-      list.remove(this);      // remove this OPfunction
-    } catch (Throwable thr) {
+      return c1w;
+    } catch (Exception thr) {
+      // if can't evaluate -- replace operands at least
+
+      if (!ep1)
+        chi[0]=new OPload(chi[0],c1w);
+      
+      if (!ep2)
+        chi[1]=new OPload(chi[1],c2w);
+
+      throw thr;
       //      Debug.reportThrowable(thr);
       // IGNORE
     };
   };
 
-  public String toString() {
-    return opSymbols[code];
-  };
-
 };
 
-
-
-
-
