@@ -1,5 +1,5 @@
 /* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- 
- * $Id: OPunary.java,v 1.11 2001/08/23 14:30:21 metlov Exp $
+ * $Id: OPunary.java,v 1.19 2002/02/13 02:14:44 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -15,13 +15,19 @@
 package gnu.jel;
 
 import gnu.jel.debug.Debug;
+import java.util.Stack;
 
-public class OPunary extends OPfunction {
-  protected int code; // operation code
+/**
+ * A tree node, representing unary operation.
+ */
+public class OPunary extends OP {
 
-  protected int uwrpCode=0; // what to code to unwrap the object
-  protected int implCode=0;
-  protected int uwrpsTo=-1;
+  /** code of this operation */
+  public int code;
+
+  private int uwrpCode=0; // what to code to unwrap the object
+  private int implCode=0;
+  private int uwrpsTo=-1;
 
   // The full list of codes is :
   // 0  -- negation (applicable to anything except boolean)
@@ -41,23 +47,20 @@ public class OPunary extends OPfunction {
   // 14 -- convert string to temporary string buffer (TSB)
   // 15 -- convert temporary string buffer (TSB) to string
 
+  /** unary promotions of base types */
   protected final static byte[] unary_prmtns;
   
   private final static int[][] una;
 
-  private final static String[] opSymbols;
-  
   private final static String[] opNames;
 
   static {
     unary_prmtns=(byte[])TableKeeper.getTable("unary_prmtns");
     una=(int[][])TableKeeper.getTable("una");
-    opSymbols=(String[])TableKeeper.getTable("opSymbols");
     opNames=(String[])TableKeeper.getTable("opNames");
 
     if (Debug.enabled)
-      Debug.assert((opNames.length==opSymbols.length) &&
-                   (opNames.length==una.length));
+      Debug.assert((opNames.length==una.length));
   };
 
   /**
@@ -69,78 +72,80 @@ public class OPunary extends OPfunction {
    * 2  -- logical not (applicable to booleans only)
    * 3  -- return the type in stack
    * </PRE>
-   * @param typesStk stack holding the current types (will be updated)
+   * @param paramOPs stack holding the operands
    * @param code operation code
-   * @param cls class to convert to (used only if code==11)
    */
-  public OPunary(TypesStack typesStk, int code) 
+  public OPunary(Stack paramOPs, int code) 
     throws CompilationException {
     if (Debug.enabled)
       Debug.assert((code>=0) && (code<=3));
 
     this.code=code;
+    chi=new OP[1];
+    chi[0]=(OP)paramOPs.pop();
     
-    int opID=typesStk.peekID();
-    Class opType=typesStk.pop();
+    int opID=chi[0].resID;
+    Class opType=chi[0].resType;
 
     if (code==3) { // return
       // everything can be returned no checks/unwraps needed
-      resID=TypesStack.baseType[opID];
+      resID=(opID>=8?8:opID);    // computes base type
+      resType=opType;
       implCode=una[code][resID];
     } else {
       
-      // unwrap object if needed
-      int unwrpID=TypesStack.unwrapType[opID];      
+      // unwrap object if can
+      int unwrpID=unwrapType[opID];      
       if (unwrpID!=opID) {
         uwrpCode=((opID-12+11)<<8)+0x00FE;
         uwrpsTo=unwrpID;
       };
-
+      
       if ((implCode=una[code][unwrpID])==0xFF) {
         // operation is not defined on types
         Object[] paramsExc={opNames[code],opType};
-        throw new CompilationException(-1,28,paramsExc);
+        throw new CompilationException(28,paramsExc);
       };
       
       resID=unary_prmtns[unwrpID];
-      typesStk.pushID(resID,null);      
+      resType=specialTypes[resID];
     };
+    
   };
 
   /**
    * Creates conversion operation to the given class.
-   * @param typesStk stack holding the current types (will be updated)
+   * @param paramOPs stack holding the operands
    * @param targetID ID of primitive type to convert to.
    * @param targetClass the class to convert to, in case cldID=8
    * @param allownarrowing if narrowing conversions are allowed.
    */
-  public OPunary(TypesStack typesStk, int targetID, Class targetClass,
+  public OPunary(Stack paramOPs, int targetID, Class targetClass,
                  boolean allownarrowing) throws CompilationException {
     if (Debug.enabled) {
       // check for proper target type was identification
       Debug.assert(((targetID==8)^(targetClass==null)) ||
                    ((targetID!=8)&&
-  (TypesStack.specialTypes[targetID].isAssignableFrom(targetClass)))
+            (specialTypes[targetID].isAssignableFrom(targetClass)))
                    );
       
       if (targetID==8) {
-        typesStk.push(targetClass);
-        Debug.assert(typesStk.peekID()==targetID,
+        Debug.assert(typeID(targetClass)==targetID,
                      "The type was improperly identified for OPunary conv.");
-        typesStk.pop();
       };
     };
-    
+
     // set the result type
     resID=targetID;
-    resType=(targetClass!=null?targetClass:TypesStack.specialTypes[targetID]);
+    resType=(targetClass!=null?targetClass:specialTypes[targetID]);
 
-    Class currClass=typesStk.peek();
-    int   currID=typesStk.peekID();
+    chi=new OP[1];
+    chi[0]=(OP)paramOPs.pop();
 
-    int unwrappedCurrID=TypesStack.unwrapType[currID];
-    int baseCurrID=TypesStack.baseType[currID];
-    int baseResID=TypesStack.baseType[resID];
+    Class currClass=chi[0].resType;
+    int   currID=chi[0].resID;
+
+    int unwrappedCurrID=unwrapType[currID];
 
     // there are folowing cases:
     // unwrappable object to primitive
@@ -152,7 +157,7 @@ public class OPunary extends OPfunction {
     // error (object -> primitive, primitive->object)
 
     code=4+resID;
-    if ((baseResID==8) && (resID!=10) && (currID!=10) && 
+    if ((resID>=8) && (resID!=10) && (currID!=10) && 
         (!((currID==28) && ((resID==10) || (resID==11))))) {
       // object to object
       code=4+8;
@@ -172,40 +177,18 @@ public class OPunary extends OPfunction {
       //Debug.println("code="+code);
       // can't convert at all
       Object[] paramsExc={currClass,resType};
-      throw new CompilationException(-1,21,paramsExc);
+      throw new CompilationException(21,paramsExc);
     };
 
     if (!(allownarrowing || 
-          TypesStack.isWidening(currID,currClass,resID,resType))) {
+          isWidening(currID,currClass,resID,resType))) {
       // can't do narrowing conversions automatically
       Object[] paramsExc={currClass,resType};
-      throw new CompilationException(-1,22,paramsExc);
+      throw new CompilationException(22,paramsExc);
     };
-    
-    typesStk.pop();
-    typesStk.pushID(resID,resType);
   };
 
-//    /**
-//     * Creates conversion operation to the given class.
-//     * @param typesStk stack holding the current types (will be updated)
-//     * @param cls the class to convert to, in case cldID=8
-//     * @param allownarrowing if narrowing conversions are allowed.
-//     */
-//    public OPunary(TypesStack typesStk, Class cls,
-//                   boolean allownarrowing) throws CompilationException {
-//      this(typesStk,TypesStack.typeID(cls),cls,allownarrowing);
-//    };
-
-
-  /**
-   * Returns number of parameters for this function.
-   */
-  public int getNParams() {
-    return 1;
-  };
-
-  protected void compile_pre(ClassFile cf) {
+  public void compile(ClassFile cf) {
     if (code==2) cf.code(0xFB); // equivalent to cf.labels_block();
     if (code==14) {
       cf.code(0x0001FE591DFDBBL); // ANY => TSB
@@ -213,22 +196,20 @@ public class OPunary extends OPfunction {
       //                          <CP: java.lang.StringBuffer>
       //                      | dup
       //                      | invokespecial StringBuffer()
-      cf.typesStk.pushID(10); // pushes ref to TSB
+      cf.noteStk(-1,10); // pushes ref to TSB
     };
-  };
-
-  protected void compile(ClassFile cf) {
+    
+    chi[0].compile(cf);
+    
     cf.code(uwrpCode); // unwrap object if needed
     
-    if (uwrpsTo>=0) { // note the unwrapping on stack
-      cf.typesStk.pop();
-      cf.typesStk.pushID(uwrpsTo);
-    };
+    if (uwrpsTo>=0)
+      cf.noteStk(chi[0].resID,uwrpsTo); // note the unwrapping on stack
 
     cf.code(implCode);
 
     if (code==12) // code CP index for conversion to reference
-      cf.codeI(cf.getIndex(resType,9));
+      cf.writeShort(cf.getIndex(resType,9));
     
     // get rid of this switch is the task for the NEXT major update
     switch(code) {
@@ -238,41 +219,40 @@ public class OPunary extends OPfunction {
       break;
     case 3:
     case 13:
-      cf.typesStk.pop(); // return and (void) throw one word, add nothing
+      cf.noteStk(resID,-1); // return and (void) throw one word, add nothing
+      break;
+    case 14:
+      cf.noteStk(11,-1);
       break;
     case 1: // bitwise not may have excess item on stack
-      cf.typesStk.tempExcessWords(TypesStack.stkoccup[resID]);
+      cf.noteStk(-1,resID);
+      cf.noteStk(resID,-1);
     default: // other ops throw one word replace it by another
-      cf.typesStk.pop();
-      cf.typesStk.pushID(resID,resType);
+      cf.noteStk(uwrpsTo>=0?uwrpsTo:chi[0].resID,resID);
     };
-
   };
 
-  /**
-   * Attempts to perform this operation.
-   * @param list is the list of OPs this one belong to, 
-   *             if eval is unsuccessful this list is not modified.
-   */
-  protected void eval(OPlist list) {
-    if (! (prev instanceof OPload)) return; // operand is not ready
+  public Object eval() throws Exception {
+    Object operand=chi[0].eval();
+    int operand_resID=chi[0].resID;
 
-    if (code==3) return; // "return" operation can't be evaluated
-    if (code==13) return; // "void" operation can't be evaluated
-    if (code==12) return; // "reference cast" operation can't be evaluated
-
-    OPload opl=(OPload) prev;
+    if ((code==3) ||
+        (code==13) ||
+        (code==12)) { // do not evaluate, just replace operand
+      chi[0]=new OPload(chi[0],operand);
+      throw new Exception(); // bail out
+    };
 
     if (code==2) { // logical not
-      if (((Boolean)opl.what).booleanValue())
-        opl.what=Boolean.FALSE;
+      if (((Boolean)operand).booleanValue())
+        operand=Boolean.FALSE;
       else
-        opl.what=Boolean.TRUE;
+        operand=Boolean.TRUE;
     } else if (code<2) {
-      Number val=TypesStack.widen(opl.what,opl.resID);
+      Number val=widen(operand,operand_resID);
       switch(code) {
       case 0:  // negation
-        if (opl.resID>5)
+        if (operand_resID>5)
           val=new Double(-val.doubleValue());
         else
           val=new Long(-val.longValue());
@@ -284,26 +264,18 @@ public class OPunary extends OPfunction {
         if (Debug.enabled)
           Debug.assert(code>=0,"Wrong unary opcode.");
       };
-      opl.what=TypesStack.narrow(val,resID);
-      opl.resID=resID;
+      operand=narrow(val,resID);
     } else {
       // conversion operations
       if (code==14) { // ANY->TSB
-        opl.what=new StringBuffer(String.valueOf(opl.what));
+        operand=new StringBuffer(String.valueOf(operand));
       } else if (code==15) { // TSB->STR
-        opl.what=opl.what.toString();
+        operand=operand.toString();
       } else {
-        opl.what=TypesStack.narrow(TypesStack.widen(opl.what,opl.resID),
-                                   resID);
+        operand=narrow(widen(operand,operand_resID),resID);
       };
-      opl.resID=resID;
     };
-    list.remove(this);
+    return operand;
   };
-
-  public String toString() {
-    return opSymbols[code];
-  };
-  
 };
 
