@@ -1,5 +1,5 @@
 /*-*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
- * $Id: VariableProvider.java,v 1.16 2001/06/08 15:48:59 metlov Exp $
+ * $Id: VariableProvider.java,v 1.17 2001/11/07 22:37:07 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -14,7 +14,7 @@
 
 package gnu.jel.tests;
 
-public class VariableProvider implements gnu.jel.DVResolver {
+public class VariableProvider extends gnu.jel.DVMap {
   public Object anObject=new Object();
 
   public double xvar;
@@ -49,12 +49,16 @@ public class VariableProvider implements gnu.jel.DVResolver {
   };
 
   public String getTypeName(String name) {
+    // this is for testing translated names
+    if (name.startsWith("_T_")) return "String";
+
     Object val=properties.get(name);
     if (val==null) return null;
     if (val instanceof String) return "String";
     if (val.getClass()==DoubleObject.class) return "TSDouble";
     if (val instanceof gnu.jel.reflect.Double) return "Double";
     if (val instanceof gnu.jel.reflect.Boolean) return "Boolean";
+    
     return null;
   };
   
@@ -235,6 +239,19 @@ public class VariableProvider implements gnu.jel.DVResolver {
     return str+"_ttt";
   };
 
+  public Object translate(String name) {
+    if (name.startsWith("_T_")) {
+      // this is for testing translated names
+      return new Integer(name.charAt(3)-'a');
+    } else
+      return name;
+  };
+
+  public String getStringProperty(int num) {
+    char lett=(char)('a'+num);
+    return "_U_"+lett;
+  };
+  
 };
 
 
